package calendar;

import java.util.Calendar;
import java.util.Date;
import java.util.Hashtable;
import java.util.TimeZone;
import java.util.Vector;

import log.ILogger;
import log.Logger;
import calendar.objectmodel.ICalendar;
import calendar.objectmodel.IEventDescription;
import calendar.objectmodel.IRepeatRule;
import calendar.objectmodel.ImmutableCalendar;
import calendar.objectmodel.ImmutableEvent;
import calendar.objectmodel.ImmutableRepeatRule;

/**
 * Primitive partial implementation of an ICal file parser.
 * 
 * Supports a minority of RFC 2445, but enough for my timetable. Includes multiple
 * calendars, events (start, end, description, location etc), limited and probably
 * buggy timezone support, and half implemented RRULE (repeat rule) support.
 * 
 * @author richard
 *
 */
public class ICalCalendarParser implements ICalendarParser {

	private final static ILogger log = Logger.get(ICalCalendarParser.class);

	private static final String EVENT_DESCRIPTION = "DESCRIPTION";

	private static final String CALENDAR_BEGIN = "BEGIN:VCALENDAR";

	private static final String CALENDAR_END = "END:VCALENDAR";

	private static final String EVENT_BEGIN = "BEGIN:VEVENT";

	private static final String EVENT_END = "END:VEVENT";

	private static final String EVENT_START_DATETIME = "DTSTART";

	private static final String EVENT_END_DATETIME = "DTEND";

	private static final String EVENT_TITLE = "SUMMARY";

	private static final String EVENT_LOCATION = "LOCATION";

	private static final String REPEAT_RULE = "RRULE";

	private static final String RULE_FREQUENCY = "FREQ";

	private static final String RULE_UNTIL = "UNTIL";

	private static final String RULE_INTERVAL = "INTERVAL";

	private static final String RULE_DAY = "BYDAY";

	private static final Hashtable REPEAT_FREQUENCY_NAMES = new Hashtable();

	private static final Hashtable REPEAT_DAY_NAMES = new Hashtable();

	private static final int REPEAT_DAY_NAMES_LENGTH = 2;

	static {
		REPEAT_FREQUENCY_NAMES.put("WEEKLY", new Integer(IRepeatRule.FREQUENCY_WEEKLY));
		REPEAT_FREQUENCY_NAMES.put("MONTHLY", new Integer(IRepeatRule.FREQUENCY_MONTHLY));
		REPEAT_FREQUENCY_NAMES.put("DAILY", new Integer(IRepeatRule.FREQUENCY_DAILY));

		REPEAT_DAY_NAMES.put("MO", new Integer(IRepeatRule.DAY_MONDAY));
		REPEAT_DAY_NAMES.put("TU", new Integer(IRepeatRule.DAY_TUESDAY));
		REPEAT_DAY_NAMES.put("WE", new Integer(IRepeatRule.DAY_WEDNESDAY));
		REPEAT_DAY_NAMES.put("TH", new Integer(IRepeatRule.DAY_THURSDAY));
		REPEAT_DAY_NAMES.put("FR", new Integer(IRepeatRule.DAY_FRIDAY));
		REPEAT_DAY_NAMES.put("SA", new Integer(IRepeatRule.DAY_SATURDAY));
		REPEAT_DAY_NAMES.put("SU", new Integer(IRepeatRule.DAY_SUNDAY));
	}

	// This must go after all the static final variables
	public static final ICalCalendarParser INSTANCE = new ICalCalendarParser();

	private ICalCalendarParser() {

	}

	public ICalendar parse(IRawCalendar rawCalendar) {
		final String[] wrappedLines = OldJavaUtils.splitIntoLines(rawCalendar.getData());
		final String[] lines = removeWrapping(wrappedLines);

		final Vector calendars = new Vector();

		final int[] newOffset = new int[1];

		for (int offset = 0; offset < lines.length; offset++) {
			if (lines[offset].startsWith(CALENDAR_BEGIN)) {
				calendars.addElement(parseCalendar(rawCalendar.getAsOf(), lines, offset, newOffset));
				offset = newOffset[0];
			}
		}

		return new CombiningCalendar((ICalendar[]) OldJavaUtils.toArray(calendars, new ICalendar[calendars.size()]));
	}

	// Undoes the wrapping which my evolution puts into the files. TODO:
	// remove escape characters too.
	private static String[] removeWrapping(String[] originalLines) {
		final Vector newLines = new Vector();

		String lineUnderConstruction = originalLines[0];

		for (int originalLineIndex = 1; originalLineIndex < originalLines.length; originalLineIndex++) {
			if (originalLines[originalLineIndex].startsWith(" ")) {
				lineUnderConstruction = lineUnderConstruction + originalLines[originalLineIndex].substring(1);
			} else {
				newLines.addElement(lineUnderConstruction);
				lineUnderConstruction = originalLines[originalLineIndex];
			}
		}
		newLines.addElement(lineUnderConstruction);

		return (String[]) OldJavaUtils.toArray(newLines, new String[newLines.size()]);
	}

	public static ICalendar parseCalendar(final Date asOf, final String[] lines, final int offset, final int[] newOffset) {
		int localOffset;
		final int[] newLocalOffset = new int[1];

		final Vector events = new Vector();

		for (localOffset = offset; localOffset < lines.length; localOffset++) {
			if (lines[localOffset].startsWith(EVENT_BEGIN)) {
				final IEventDescription newEvent = parseEvent(lines, localOffset, newLocalOffset);
				if (newEvent != null)
					events.addElement(newEvent);
				localOffset = newLocalOffset[0];
			}
			if (lines[offset].startsWith(CALENDAR_END)) {
				break;
			}
		}

		newOffset[0] = localOffset;
		return new ImmutableCalendar(asOf, (IEventDescription[]) OldJavaUtils.toArray(events,
				new IEventDescription[events.size()]));
	}

	private static String extractParameter(String line, char start, char end) {
		return line.substring(line.indexOf(start) + 1, line.indexOf(end));
	}

	public static IEventDescription parseEvent(String[] lines, int offset, int[] newOffset) {
		int localOffset;

		Date start = null, end = null;
		String title = null, description = null;
		IRepeatRule repeatRule = null;
		String location = null;

		for (localOffset = offset; localOffset < lines.length; localOffset++) {
			final String line = lines[localOffset];

			if (line.startsWith(EVENT_START_DATETIME)) {
				final String parameter = extractParameter(line, ';', ':');
				start = parseDate(line, line.indexOf(":") + 1, parameter);
			} else if (line.startsWith(EVENT_END_DATETIME)) {
				final String parameter = extractParameter(line, ';', ':');
				end = parseDate(line, line.indexOf(":") + 1, parameter);
			} else if (line.startsWith(EVENT_TITLE)) {
				title = new String(line.substring(line.indexOf(":") + 1));
			} else if (line.startsWith(EVENT_DESCRIPTION)) {
				description = new String(line.substring(line.indexOf(":") + 1));
			} else if (line.startsWith(REPEAT_RULE)) {
				repeatRule = parseRepeatRule(line);
			} else if (line.startsWith(EVENT_LOCATION)) {
				location = new String(line.substring(line.indexOf(":") + 1));
			} else if (line.startsWith(EVENT_END)) {
				break;
			}
		}
		newOffset[0] = localOffset;

		if (title == null || start == null || end == null) {
			final String error = "Could not read from " + offset + " to " + localOffset;
			log.log(error, ILogger.ERROR);
			return null;
		} else {
			return new ImmutableEvent(title, description, location, start, end, repeatRule);
		}
	}

	private static IRepeatRule parseRepeatRule(String line) {
		final String parameters = line.substring(line.indexOf(":") + 1);
		final String[] sections = OldJavaUtils.split(parameters, ";");

		Date until = null;
		Integer repeatFrequency = null;
		Integer repeatInterval = null;
		Integer indexOfDayInMonthOrYear = null;
		int repeatDay = 0;

		for (int sectionIndex = 0; sectionIndex < sections.length; sectionIndex++) {
			final String section = sections[sectionIndex];

			if (section.startsWith(RULE_UNTIL)) {
				until = parseDate(section, section.indexOf("=") + 1, null);
			} else if (section.startsWith(RULE_FREQUENCY)) {
				final String frequencyName = section.substring(section.indexOf("=") + 1);
				repeatFrequency = ((Integer) REPEAT_FREQUENCY_NAMES.get(frequencyName));
			} else if (section.startsWith(RULE_INTERVAL)) {
				repeatInterval = Integer.valueOf(section.substring(section.indexOf("=") + 1));
			} else if (section.startsWith(RULE_DAY)) {
				final String dayList = section.substring(section.indexOf("=") + 1);
				final String[] dayNames = OldJavaUtils.split(dayList, ",");
				for (int dayIndex = 0; dayIndex < dayNames.length; dayIndex++) {
					final String dayName = dayNames[dayIndex];
					final String trimmedDayName;
					if (dayName.length() > REPEAT_DAY_NAMES_LENGTH) {
						final String indexOfDayString = dayName
								.substring(0, dayName.length() - REPEAT_DAY_NAMES_LENGTH);
						indexOfDayInMonthOrYear = Integer.valueOf(indexOfDayString);
						trimmedDayName = dayName.substring(dayName.length() - REPEAT_DAY_NAMES_LENGTH);
					} else {
						trimmedDayName = dayName;
					}
					repeatDay += ((Integer) REPEAT_DAY_NAMES.get(trimmedDayName)).intValue();
				}
			}
		}
		return new ImmutableRepeatRule(repeatFrequency.intValue(), until, repeatDay);
	}

	private static Date parseDate(String line, int offset, String parameter) {

		final int year = Integer.parseInt(line.substring(offset, offset + 4));
		final int month = Integer.parseInt(line.substring(offset + 4, offset + 4 + 2));
		final int day = Integer.parseInt(line.substring(offset + 6, offset + 6 + 2));

		final boolean forceAsDate = "VALUE=DATE".equals(parameter);
		
		final int hour, minute, second;

		if (!forceAsDate && offset + 13 + 2 <= line.length()) {
			hour = Integer.parseInt(line.substring(offset + 9, offset + 9 + 2));
			minute = Integer.parseInt(line.substring(offset + 11, offset + 11 + 2));
			second = Integer.parseInt(line.substring(offset + 13, offset + 13 + 2));
		} else {
			hour = 0;
			minute = 0;
			second = 0;
		}

		final int possibleZCharOffset = offset + 8 + 1 + 6;
		final boolean isUTC = (possibleZCharOffset) < line.length() && line.charAt(possibleZCharOffset) == 'Z';

		final TimeZone timezone;

		if (isUTC) {
			timezone = TimeZonesManager.UTC;
		} else {
			if (parameter != null) {
				final String timezonePrefix = "TZID=";
				if (parameter.startsWith(timezonePrefix))
					timezone = TimeZonesManager.get(parameter.substring(timezonePrefix.length()));
				else
					timezone = TimeZone.getDefault();
			} else {
				timezone = TimeZone.getDefault();
			}
		}
		
		log.log("Parse [" + line + "] as " + hour + " " + isUTC, ILogger.DEBUG);

		final Calendar calendar = Calendar.getInstance(timezone);
		calendar.set(Calendar.DAY_OF_MONTH, day);
		calendar.set(Calendar.MONTH, month - 1);
		calendar.set(Calendar.YEAR, year);
		calendar.set(Calendar.HOUR_OF_DAY, hour);
		calendar.set(Calendar.MINUTE, minute);
		calendar.set(Calendar.SECOND, second);
		final Date newTime = calendar.getTime();

		log.log("Parse [" + line + "] as " + newTime, ILogger.DEBUG);
		return newTime;
	}
}
