package calendar.objectmodel;

import java.util.Calendar;
import java.util.Date;
import java.util.Hashtable;
import java.util.Vector;

import log.ILogger;

import calendar.OldJavaUtils;
import calendar.RepeatEvent;
import calendar.SingleInstanceEvent;

/**
 * A standard ICalendar implementation, which interprets event series (IEventDescriptions)
 * and produces concrete events (IEventInstance). They are different in that a single
 * IEventDescription may correspond to an infinite number of IEventInstances for series
 * that repeat forever. 
 * 
 * This class therefore implements the repeat rules.
 * 
 * @author richard
 *
 */
public class ImmutableCalendar implements ICalendar {

	private final IEventDescription[] events;
	private final Date asOf;

	public ImmutableCalendar(Date asOf, IEventDescription[] events) {
		this.asOf = asOf;
		this.events = events;
	}

	public IEventInstance[] getEvents(final Date date, ILogger logger) {

		final Vector validEvents = new Vector();

		for (int eventDescriptionIndex = 0; eventDescriptionIndex < events.length; eventDescriptionIndex++) {
			final IEventDescription eventDescription = events[eventDescriptionIndex];

			final Date dayOfEvent = OldJavaUtils.stripTime(eventDescription.getStart());
			if (dayOfEvent.equals(date)) {
				validEvents.addElement(new SingleInstanceEvent(eventDescription));
			} else {
				// Check repeating events
				final IRepeatRule repeatRule = eventDescription.getRepeatRule();
				if (repeatRule != null) {

					if (dayOfEvent.getTime() <= date.getTime()
							&& (repeatRule.getUntil() == null || repeatRule.getUntil().getTime() >= date.getTime())) {
						final boolean addEventToday;

						logger.log("Event still running " + eventDescription, ILogger.DEBUG);

						if (repeatRule.getFrequency() == IRepeatRule.FREQUENCY_DAILY) {
							addEventToday = true;
						} else if (repeatRule.getFrequency() == IRepeatRule.FREQUENCY_WEEKLY) {
							final Calendar calendar = Calendar.getInstance();
							calendar.setTime(date);
							int dayOfWeekRequested = calendar.get(Calendar.DAY_OF_WEEK);

							logger.log("Day of week requested = " + dayOfWeekRequested, ILogger.DEBUG);

							if (daysMatch(dayOfWeekRequested, repeatRule.getDays()))
								addEventToday = true;
							else
								addEventToday = false;
						} else
							addEventToday = false;

						if (addEventToday) {
							final Date newStart = createNewDateAtSameTime(eventDescription.getStart(), date);
							validEvents.addElement(new RepeatEvent(eventDescription, newStart));
						}
					}
				}
			}
		}

		return (IEventInstance[]) OldJavaUtils.toArray(validEvents, new IEventInstance[validEvents.size()]);
	}

	private static Hashtable CALENDAR_TO_REPEAT_RULE_DAYS_OF_WEEK = new Hashtable();

	static {
		recordDayMatch(Calendar.MONDAY, IRepeatRule.DAY_MONDAY);
		recordDayMatch(Calendar.TUESDAY, IRepeatRule.DAY_TUESDAY);
		recordDayMatch(Calendar.WEDNESDAY, IRepeatRule.DAY_WEDNESDAY);
		recordDayMatch(Calendar.THURSDAY, IRepeatRule.DAY_THURSDAY);
		recordDayMatch(Calendar.FRIDAY, IRepeatRule.DAY_FRIDAY);
		recordDayMatch(Calendar.SATURDAY, IRepeatRule.DAY_SATURDAY);
		recordDayMatch(Calendar.SUNDAY, IRepeatRule.DAY_SUNDAY);
	}

	private static void recordDayMatch(int calendarDayOfWeek, int repeatDayOfWeek) {
		CALENDAR_TO_REPEAT_RULE_DAYS_OF_WEEK.put(new Integer(calendarDayOfWeek), new Integer(repeatDayOfWeek));
	}

	private static boolean daysMatch(int calendarDayOfWeek, int repeatDaysOfWeek) {
		System.out.println("Event runs on days " + repeatDaysOfWeek);
		final int repeatDayOfWeek = ((Integer) CALENDAR_TO_REPEAT_RULE_DAYS_OF_WEEK.get(new Integer(calendarDayOfWeek)))
				.intValue();
		System.out.println("Translated calendar day " + repeatDayOfWeek);
		return (repeatDayOfWeek & repeatDaysOfWeek) == repeatDayOfWeek;
	}

	private static Date createNewDateAtSameTime(Date oldTime, Date oldDate) {
		final Calendar calendar = Calendar.getInstance();
		calendar.setTime(oldTime);

		int hour = calendar.get(Calendar.HOUR_OF_DAY);
		int minute = calendar.get(Calendar.MINUTE);
		int second = calendar.get(Calendar.SECOND);

		calendar.setTime(oldDate);
		calendar.set(Calendar.HOUR_OF_DAY, hour);
		calendar.set(Calendar.MINUTE, minute);
		calendar.set(Calendar.SECOND, second);
		return calendar.getTime();
	}

	public Date getAsOf() {
		return asOf;
	}
}
