package calendarsourceentries;

import log.ILogger;
import calendar.ICalendarSource;
import calendar.OldJavaUtils;

/**
 * Represents a Calendar configured by the user, which can be saved in the phone
 * storage between each use of the program.
 * 
 * These SourceEntries can create ICalendarSource objects, so this class acts as a factory
 * for the various implementations of sources of calendars.
 * 
 * @author richard
 *
 */
public abstract class CalendarSourceEntry {
	private final String name;

	private static final String DATA_SEPERATOR = "|";

	public CalendarSourceEntry(String name) {
		this.name = name;
	}

	protected abstract String[] internalSerialize();

	protected abstract String getMarker();

	public abstract ICalendarSource[] createSources(ILogger log);

	public String getName() {
		return name;
	}

	public static CalendarSourceEntry parse(byte[] data) {
		final String stringData = new String(data);
		final String[] parameters = OldJavaUtils.split(stringData, DATA_SEPERATOR);

		final String name = parameters[1];
		
		final String[] typeParameters = new String[parameters.length - 2];
		System.arraycopy(parameters, 2, typeParameters, 0, typeParameters.length);

		if (parameters[0].equals(CalendarFileSourceEntry.MARKER))
			return CalendarFileSourceEntry.parse(name, typeParameters);
		if (parameters[0].equals(CalendarHttpSourceEntry.MARKER))
			return CalendarHttpSourceEntry.parse(name, typeParameters);
		else
			throw new RuntimeException("Unrecognised calendar type " + parameters[0]);
	}

	public byte[] serialize() {
		final String[] typeParameters = internalSerialize();

		final String[] fullParameters = new String[typeParameters.length + 2];
		fullParameters[0] = getMarker();
		fullParameters[1] = name;
		System.arraycopy(typeParameters, 0, fullParameters, 2, typeParameters.length);

		int size = 0;
		for (int i = 0; i < fullParameters.length; i++)
			size += fullParameters[i].length() + DATA_SEPERATOR.length();

		byte[] toReturn = new byte[size];
		int index = 0;
		for (int i = 0; i < fullParameters.length; i++) {
			final byte[] parameterBytes = (fullParameters[i] + DATA_SEPERATOR).getBytes();
			System.arraycopy(parameterBytes, 0, toReturn, index, parameterBytes.length);
			index += parameterBytes.length;
		}
		return toReturn;
	}
}
