package phonecalendar;

import java.util.Date;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Vector;

import javax.microedition.lcdui.Command;
import javax.microedition.lcdui.CommandListener;
import javax.microedition.lcdui.DateField;
import javax.microedition.lcdui.Display;
import javax.microedition.lcdui.Displayable;
import javax.microedition.lcdui.Form;
import javax.microedition.lcdui.Gauge;
import javax.microedition.lcdui.StringItem;

import log.ILogger;
import phonecalendar.CalendarList.CalendarsChangedListener;
import calendar.ICalendarSource;
import calendar.OldJavaUtils;
import calendar.ICalendarSource.IObserver;
import calendar.objectmodel.ICalendar;
import calendar.objectmodel.IEventInstance;
import calendarsourceentries.CalendarSourceHolder;

public class MainForm extends Form {

	private final Display display;

	// of CalendarSourceEntry
	private Vector knownCalendars;

	// Of LoadedCalendar
	private Vector loadedCalendars = new Vector();

	// Of CalendarSourceEntry to LoadingCalendars
	private Vector loadingCalendars = new Vector();

	private final CalendarList calendarList;

	private final ILogger log;

	private final DateField calendarControl;

	public MainForm(final Display display, final ILogger log, final Vector knownCalendars) {
		super("Select Date");

		this.knownCalendars = knownCalendars;
		this.display = display;
		this.log = log;

		final FormClosedListener bringBackToUs = new FormClosedListener() {
			public void closed(int result) {
				display.setCurrent(MainForm.this);
			}
		};

		// Event list
		final EventList eventList = new EventList(display);
		eventList.setFormClosedListener(bringBackToUs);

		final CalendarsChangedListener calendarsListener = new CalendarsChangedListener() {
			public void deleted(CalendarSourceHolder calendar) {
				removeLoadedForCalendar(calendar);
				buildUI();
			}

			public void updatedOrCreated(CalendarSourceHolder calendar) {
				startLoadingCalendar(log, calendar);
			}
		};

		// CalendarList
		calendarList = new CalendarList(calendarsListener, display);
		calendarList.setFormClosedListener(bringBackToUs);

		calendarControl = new DateField("", DateField.DATE);
		calendarControl.setDate(OldJavaUtils.stripTime(new Date()));

		final Command viewDateCommand = new Command("Events", Command.OK, 1);
		final Command viewCalendarsCommand = new Command("Calendars", Command.SCREEN, 1);
		addCommand(viewDateCommand);
		addCommand(viewCalendarsCommand);
		setCommandListener(new CommandListener() {
			public void commandAction(Command command, Displayable arg1) {
				System.out.println("Command " + command);
				if (command == viewDateCommand) {
					System.out.println("View date clicked");
					final Date date = calendarControl.getDate();
					final IEventInstance[] eventsForDay = getEventsForDay(date);
					eventList.populateEventList(eventsForDay, date);
					display.setCurrent(eventList);
				} else if (command == viewCalendarsCommand) {
					populateCalendarList();
					display.setCurrent(calendarList);
				}
			}
		});

		startLoadingKnownCalendars(log);
	}

	private void buildUICrossThread() {
		display.callSerially(new Runnable() {
			public void run() {
				buildUI();
			}
		});
	}

	private void buildUI() {
		log.log("Building UI", ILogger.INFO);
		deleteAll();

		synchronized (this) {
			log.log("There are " + loadedCalendars.size() + " calendars", ILogger.INFO);

			append(calendarControl);
			for (int i = 0; i < loadedCalendars.size(); i++) {
				final LoadedCalendar calendar = (LoadedCalendar) loadedCalendars.elementAt(i);
				calendar.appendUI();
			}

			log.log("There are " + loadingCalendars.size() + " loading calendars", ILogger.INFO);

			for (int i = 0; i < loadingCalendars.size(); i++) {
				final LoadingCalendar calendar = (LoadingCalendar) loadingCalendars.elementAt(i);
				calendar.appendUI();
			}
		}
		log.log("Finished building UI", ILogger.INFO);
	}

	private IEventInstance[] getEventsForDay(Date date) {
		final Hashtable calendarForSource = new Hashtable();

		synchronized (this) {
			for (Enumeration iterator = loadedCalendars.elements(); iterator.hasMoreElements();) {
				final LoadedCalendar loadedCalendar = (LoadedCalendar) iterator.nextElement();
				final LoadedCalendar previousSuggestedCalendar = (LoadedCalendar) calendarForSource.get(loadedCalendar
						.getCalendarHolder());

				// Prioritise one over the other
				if (!(loadedCalendar.getCalendar() == null && (previousSuggestedCalendar == null || previousSuggestedCalendar
						.getCalendar() == null))) {
					// By this point at least one calendar is valid
					final LoadedCalendar selectedCalendar;
					// If one is invalid, always take the other
					if (loadedCalendar == null)
						selectedCalendar = previousSuggestedCalendar;
					else if (previousSuggestedCalendar == null || previousSuggestedCalendar.getCalendar() == null)
						selectedCalendar = loadedCalendar;
					// Then prioritise based on index, later is better
					else if (previousSuggestedCalendar.getSourceIndex() > loadedCalendar.getSourceIndex())
						selectedCalendar = previousSuggestedCalendar;
					else
						selectedCalendar = loadedCalendar;
					calendarForSource.put(loadedCalendar.getCalendarHolder(), selectedCalendar);
				}
			}
		}

		final Vector events = new Vector();

		// Taking the selected calendars, fetch events
		for (Enumeration iterator = calendarForSource.elements(); iterator.hasMoreElements();) {
			final LoadedCalendar selectedCalendar = (LoadedCalendar) iterator.nextElement();
			final IEventInstance[] eventsForCalendar = selectedCalendar.getCalendar().getEvents(date, log);
			for (int i = 0; i < eventsForCalendar.length; i++)
				events.addElement(eventsForCalendar[i]);
		}

		return (IEventInstance[]) OldJavaUtils.toArray(events, new IEventInstance[events.size()]);
	}

	private void populateCalendarList() {
		calendarList.populateCalendarList(knownCalendars);
	}

	private void startLoadingKnownCalendars(ILogger log) {
		for (int i = 0; i < knownCalendars.size(); i++) {
			final CalendarSourceHolder holder = (CalendarSourceHolder) knownCalendars.elementAt(i);
			startLoadingCalendar(log, holder);
		}
	}

	private void startLoadingCalendar(final ILogger log, final CalendarSourceHolder calendar) {
		new Thread() {
			public void run() {
				synchronized (MainForm.this) {
					log.log("Starting to load for " + calendar.getEntry().getName(), ILogger.INFO);
					final ICalendarSource[] sources = calendar.getEntry().createSources(log);
					log.log("It has " + sources.length + " sources", ILogger.INFO);
					cancelLoadersForHolder(calendar);
					for (int i = 0; i < sources.length; i++) {
						final LoadingCalendar loader = new LoadingCalendar(calendar, sources[i], i);
						loadingCalendars.addElement(loader);
						new Thread(loader).start();
					}
				}
				buildUICrossThread();
			}
		}.start();
	}

	private void cancelLoadersForHolder(CalendarSourceHolder holder) {
		synchronized (this) {
			for (int i = 0; i < loadingCalendars.size(); i++) {
				final LoadingCalendar loader = (LoadingCalendar) loadingCalendars.elementAt(i);
				if (holder == loader.calendarHolder) {
					loader.cancel();
				}
			}
		}
	}

	private void removeLoadedForCalendar(CalendarSourceHolder holder) {
		synchronized (this) {
			for (int i = 0; i < loadedCalendars.size(); i++) {
				final LoadedCalendar loader = (LoadedCalendar) loadedCalendars.elementAt(i);
				if (holder == loader.calendarHolder) {
					loadedCalendars.removeElementAt(i);
					i--;
				}
			}
		}
	}

	private class LoadedCalendar {
		private final StringItem label;

		private final ICalendar calendar;

		private final CalendarSourceHolder calendarHolder;

		private final int sourceIndex;

		public LoadedCalendar(CalendarSourceHolder calendarHolder, String error, ICalendar calendar, int sourceIndex) {
			label = new StringItem(calendarHolder.getEntry().getName(), error == null ? "Loaded" : "Error: " + error);
			this.calendarHolder = calendarHolder;
			this.sourceIndex = sourceIndex;
			this.calendar = calendar;
		}

		public CalendarSourceHolder getCalendarHolder() {
			return calendarHolder;
		}

		public ICalendar getCalendar() {
			return calendar;
		}

		public int getSourceIndex() {
			return sourceIndex;
		}

		public void appendUI() {
			MainForm.this.append(label);
		}
	}

	private class LoadingCalendar implements Runnable {
		private boolean cancel = false;

		private final int sourceIndex;

		private final Gauge gauge;

		private final CalendarSourceHolder calendarHolder;

		private final ICalendarSource calendarSource;

		public LoadingCalendar(final CalendarSourceHolder calendarHolder, final ICalendarSource calendarSource,
				int sourceIndex) {
			this.calendarHolder = calendarHolder;
			this.calendarSource = calendarSource;
			this.sourceIndex = sourceIndex;
			gauge = new Gauge("Loading " + calendarHolder.getEntry().getName(), false, 100, 0);
		}

		public void appendUI() {
			MainForm.this.append(gauge);
		}

		public void cancel() {
			// Don't yet support...
		}

		private ICalendar loadCalendar() {
			log.log("Starting to load calendar from source " + calendarHolder.getEntry().getName(), ILogger.INFO);
			return calendarSource.getCalender(log, new IObserver() {
				private int previousValue = Integer.MIN_VALUE;

				public void progress(final int percent) {
					if (percent > previousValue + 5) {
						previousValue = percent;
						display.callSerially(new Runnable() {
							public void run() {
								gauge.setValue(percent);
							}
						});
					}
				}
			});
		}

		public void run() {
			try {
				ICalendar loadedCalendar = null;
				Exception error = null;

				try {
					loadedCalendar = loadCalendar();
				} catch (Exception e) {
					error = e;
				}

				// Check did we successfully load
				if (loadedCalendar != null) {
					log.log("Successfully loaded calendar " + calendarHolder.getEntry().getName(), ILogger.INFO);
				} else {
					log.log("Failed to load " + calendarHolder.getEntry().getName() + "(" + sourceIndex + "), "
							+ error.toString(), ILogger.ERROR);
				}

				Vector newLoadedCalendars = new Vector();

				synchronized (MainForm.this) {

					for (int alreadyLoadedCalendarIndex = 0; alreadyLoadedCalendarIndex < loadedCalendars.size(); alreadyLoadedCalendarIndex++) {
						final LoadedCalendar alreadyLoadedCalendar = (LoadedCalendar) loadedCalendars
								.elementAt(alreadyLoadedCalendarIndex);
						boolean keepCalendar = true;

						log.log("Should we keep " + alreadyLoadedCalendar.getCalendarHolder().getEntry().getName()
								+ "," + alreadyLoadedCalendar.getSourceIndex(), ILogger.INFO);
						log.log("We have " + calendarHolder.getEntry().getName() + "," + sourceIndex, ILogger.INFO);

						if (alreadyLoadedCalendar.getCalendarHolder() == calendarHolder
								&& alreadyLoadedCalendar.getSourceIndex() == sourceIndex) {
							if (alreadyLoadedCalendar.getCalendar() == null) {
								// Previous load failed, replace!
								keepCalendar = false;
							} else if (loadedCalendar != null
									&& alreadyLoadedCalendar.getCalendar().getAsOf().getTime() < loadedCalendar
											.getAsOf().getTime()) {
								// We have a more uptodate calendar, replace!
								keepCalendar = false;
							}
						}
						log.log(keepCalendar ? "True" : "False", ILogger.INFO);

						if (keepCalendar)
							newLoadedCalendars.addElement(alreadyLoadedCalendar);
					}

					final String errorMessage = (error != null ? error.getMessage() : null);

					newLoadedCalendars.addElement(new LoadedCalendar(calendarHolder, errorMessage, loadedCalendar,
							sourceIndex));

					loadedCalendars = newLoadedCalendars;

					loadingCalendars.removeElement(this);
				}

				display.callSerially(new Runnable() {
					public void run() {
						buildUI();
					}
				});
			} catch (Exception e) {
				throw new RuntimeException(e.getMessage());
			}
		}
	}

	public Vector getKnownCalendars() {
		return knownCalendars;
	}
}
